# Local modules
from .constants.ParticleConstants import *
from .constants.VerbEndingConstants import *
from .constants.EnumeratedTypes import Formality, Tense, VerbClass

from .Utils import *

# ---------------------------------------------------------- #
#                       Positive Verb Forms                  #
# ---------------------------------------------------------- #
class PositiveVerbForms:
    def generate_plain_form(self, verb, verb_class, tense):
        '''Generate the positive polite form of the verb depending
        on the tense.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 

        Returns:
            str: positive plain form of the verb based on the tense parameter
        '''
        if tense == Tense.NONPAST:
            return verb
        return base_te_ta_form(verb, verb_class, TA_PARTICLE, DA_PARTICLE)

    def generate_polite_form(self, verb, verb_class, tense):
        '''Generate the positive polite form of the verb depending
        on the tense.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 

        Returns:
            str: positive polite form of the verb based on the tense parameter
        '''
        ending = MASU_POSITIVE_PAST
        if tense == Tense.NONPAST:
            ending = MASU_POSITIVE_NONPAST
    
        if verb_class == VerbClass.IRREGULAR:
            # ending is the same regardless. Any way to account for this? 
            return handle_irregular_verb(verb, True, ending, ending)
        else:
            verb_stem = splice_verb(verb, verb_class)
            if verb_class == VerbClass.GODAN:
                verb_stem = map_dictionary_to_i_ending(verb)
            # ichidan verb class check is not needed here
            return "{}{}".format(verb_stem, ending)

    def generate_te_form(self, verb, verb_class):
        '''Utilize base_te_ta_form function to generate the -te form 
        of the verb

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs

        Returns:
            str: -te form of the verb
        '''
        return base_te_ta_form(verb, verb_class, TE_PARTICLE, DE_PARTICLE)

    def generate_conditional_form(self, verb, verb_class, formality):
        '''Generate the positive conditional form of the verb depending
        on the level of formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive conditional form of the verb based on the formality
        parameter
        '''
        if formality == Formality.PLAIN:
            verb = base_te_ta_form(verb, verb_class, TA_PARTICLE, DA_PARTICLE)
        else:
            verb = self.generate_polite_form(verb, verb_class, Tense.PAST)
        return "{}{}".format(verb, RA_PARTICLE)

    def generate_volitional_form(self, verb, verb_class, formality):
        '''Generate the positive volitional form of the verb depending
        on the level of formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive volitional form of the verb based on the formality
            parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            return handle_irregular_verb(verb, suru_ending=VOLITIONAL_SURU_ENDING, kuru_ending=VOLITIONAL_KURU_ENDING)
        else:
            verb_base = ""
            ending = ""
            if verb_class == VerbClass.GODAN:
                # assuming plain formality param
                verb_base = map_dictionary_to_o_ending(verb)
                ending = U_PARTICLE
                if formality == Formality.POLITE:
                    verb_base = map_dictionary_to_i_ending(verb)
                    ending = VOLITIONAL_POLITE_ENDING
            elif verb_class == VerbClass.ICHIDAN:
                verb_base = splice_verb(verb, verb_class)
                # assuming plain formality param
                ending = VOLITIONAL_ICHIDAN_PLAIN_ENDING
                if formality == Formality.POLITE:
                    ending = VOLITIONAL_POLITE_ENDING
        return "{}{}".format(verb_base, ending)

    def generate_potential_form(self, verb, verb_class, formality):
        '''Generate the positive potential form of the verb depending
        on the level of formality.
        
        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive potential form of the verb based on the specified formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=POTENTIAL_SURU_PLAIN_POSITIVE_ENDING, kuru_ending=POTENTIAL_KURU_PLAIN_POSITIVE_ENDING)
            else:
                return handle_irregular_verb(verb, suru_ending=POTENTIAL_SURU_POLITE_POSITIVE_ENDING, kuru_ending=POTENTIAL_KURU_POLITE_POSITIVE_ENDING)
        else:
            # assuming godan plain form 
            verb_base = map_dictionary_to_e_ending(verb)
            ending = RU_PARTICLE
            
            if verb_class == VerbClass.GODAN:
                if formality == Formality.POLITE:
                    verb_base = map_dictionary_to_e_ending(verb)
                    ending = MASU_POSITIVE_NONPAST
            elif verb_class == VerbClass.ICHIDAN:
                verb_base = splice_verb(verb, verb_class)
                if formality == Formality.PLAIN:
                    ending = POTENTIAL_ICHIDAN_ENDING
                else:
                    ending = POTENTIAL_POLITE_ICHIDAN_ENDING
            return "{}{}".format(verb_base, ending)

    def generate_imperative_form(self, verb, verb_class, formality):
        '''Generate the positive imperative form of the verb depending
        on the level of formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive imperative form based on the specified formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=IMPERATIVE_SURU_PLAIN_POSITIVE_ENDING, kuru_ending=IMPERATIVE_KURU_PLAIN_POSITIVE_ENDING)
            else:
                return "{}{}".format(self.generate_te_form(verb, verb_class), KUDASAI)
        else:
            verb_base = self.generate_te_form(verb, verb_class)
            ending = KUDASAI
            if formality == Formality.PLAIN:
                if verb_class == VerbClass.GODAN:
                    return map_dictionary_to_e_ending(verb)
                else:
                    verb_base = splice_verb(verb, verb_class)
                    ending = RO_PARTICLE
            return "{}{}".format(verb_base, ending)

    def generate_provisional_form(self, verb, verb_class, formality=None):
        '''Generate the positive provisional form of the verb depending
        on the level of formality. No formality parameter required for 
        non-irregular verbs.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (:obj: enum, optional): Formality Enum representing the formality class
                for the conjugated verb. Defaults to None.

        Returns:
            str: positive provisional form based on the specified formality
        parameter
        '''        
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=PROVISIONAL_SURU_PLAIN_POSITIVE_ENDING, kuru_ending=PROVISIONAL_KURU_PLAIN_POSITIVE_ENDING)
            else:
                return handle_irregular_verb(verb, suru_ending=PROVISIONAL_SURU_POLITE_POSITIVE_ENDING, kuru_ending=PROVISIONAL_KURU_POLITE_POSITIVE_ENDING)
        else:
            # assuming godan verb 
            verb_base = map_dictionary_to_e_ending(verb)
            ending = BA_PARTICLE

            if verb_class == VerbClass.ICHIDAN:
                verb_base = splice_verb(verb, verb_class)
                ending = "{}{}".format(RE_PARTICLE, BA_PARTICLE)
            return "{}{}".format(verb_base, ending)

    def generate_causative_form(self, verb, verb_class, formality):
        '''Generate the positive causative form of the verb depending
        on the level of formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive causative form based on the specified formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=CAUSATIVE_PLAIN_SURU_ENDING, kuru_ending=CAUSATIVE_PLAIN_KURU_ENDING)
        else:
            # TODO FIX ALL THIS LOGIC... GROSS :-P
            if verb_class == VerbClass.GODAN:
                verb_with_a_ending = map_dictionary_to_a_ending(verb)
                if formality == Formality.PLAIN:
                    return "{}{}{}".format(verb_with_a_ending, SE_PARTICLE, RU_PARTICLE)
                else:
                    return "{}{}{}".format(verb_with_a_ending, SE_PARTICLE, MASU_POSITIVE_NONPAST)
            elif verb_class == VerbClass.ICHIDAN:
                verb_stem = splice_verb(verb, verb_class)
                if formality == Formality.PLAIN:
                    return "{}{}{}{}".format(verb_stem, SA_PARTICLE, SE_PARTICLE, RU_PARTICLE)
                else:
                    return "{}{}{}{}".format(verb_stem, SA_PARTICLE, SE_PARTICLE, MASU_POSITIVE_NONPAST)

    def generate_passive_form(self, verb, verb_class, formality):
        '''Generate the positive passive form of the verb depending
        on the level of formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: positive passive form based on the specified formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=PASSIVE_SURU_PLAIN_POSITIVE_ENDING, kuru_ending=PASSIVE_KURU_PLAIN_POSITIVE_ENDING)
        else:
            if verb_class == VerbClass.GODAN:
                verb_with_a_ending = map_dictionary_to_a_ending(verb)
                if formality == Formality.PLAIN:
                    return "{}{}{}".format(verb_with_a_ending, RE_PARTICLE, RU_PARTICLE)
                else:
                    return "{}{}{}".format(verb_with_a_ending, RE_PARTICLE, MASU_POSITIVE_NONPAST)
            elif verb_class == VerbClass.ICHIDAN:
                verb_stem = splice_verb(verb, verb_class)
                if formality == Formality.PLAIN:
                    return "{}{}".format(verb_stem, PASSIVE_ICHIDAN_PLAIN_POSITIVE_ENDING)
                else:
                    return "{}{}".format(verb_stem, PASSIVE_ICHIDAN_POLITE_POSITIVE_ENDING)



# Local modules
from .constants.ParticleConstants import *
from .constants.VerbEndingConstants import *
from .constants.EnumeratedTypes import Formality, Tense, VerbClass

from .Utils import *

# ---------------------------------------------------------- #
#                       Negative Verb Forms                  #
# ---------------------------------------------------------- #
class NegativeVerbForms:
    def generate_plain_form(self, verb, verb_class, tense):
        '''Generate the negative plain form of the verb depending
        on the tense.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 

        Returns:
            str: negative plain form of the verb based on the tense
        parameter
        '''
        verb = generate_nai_form(verb, verb_class, True)
        if tense == Tense.NONPAST:
            return verb
        # force a non-irregular verb class to remove ending -i particle
        return "{}{}".format(splice_verb(verb, VerbClass.NONIRREGULAR), KATTA_ENDING)

    def generate_polite_form(self, verb, verb_class, tense):
        '''Generate the negative polite form of the verb depending
        on the tense.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 

        Returns:
            str: negative polite form of the verb based on the tense
        parameter
        '''
        ending = MASU_NEGATIVE_PAST
        if tense == Tense.NONPAST:
            ending = MASU_NEGATIVE_NONPAST

        if verb_class == VerbClass.IRREGULAR:
            return handle_irregular_verb(verb, append_stem_particle=True, suru_ending=ending, kuru_ending=ending)
        else:
            verb_base = splice_verb(verb, verb_class)
            if verb_class == VerbClass.GODAN:
                verb_base = map_dictionary_to_i_ending(verb)
            # no change needed for ichidan verb
            return "{}{}".format(verb_base, ending)


    def generate_conditional_form(self, verb, verb_class, formality):
        '''Generate the negative conditional form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative conditional form of the verb based on the formality
        parameter
        '''
        if formality == Formality.PLAIN:
            verb = self.generate_plain_form(verb, verb_class, Tense.PAST)
        else:
            verb = self.generate_polite_form(verb, verb_class, Tense.PAST)
        return "{}{}".format(verb, RA_PARTICLE)   

    def generate_volitional_form(self, verb, verb_class, formality):
        '''Generate the negative volitional form of the verb depending
        on the formality. 

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative volitional form of the verb based on the formality
        parameter
        '''
        verb_nai_form = generate_nai_form(verb, verb_class, True)
        if formality == Formality.PLAIN:
            return "{}{}".format(verb_nai_form, VOLITIONAL_PLAIN_COPULA)
        elif formality == Formality.POLITE:
            return "{}{}".format(verb_nai_form, VOLITIONAL_POLITE_COPULA)

    def generate_potential_form(self, verb, verb_class, formality):
        '''Generate the negative potential form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative potential form of the verb based on the formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending=POTENTIAL_SURU_PLAIN_NEGATIVE_ENDING, kuru_ending=POTENTIAL_KURU_PLAIN_NEGATIVE_ENDING)
            else:
                return handle_irregular_verb(verb, suru_ending=POTENTIAL_SURU_POLITE_NEGATIVE_ENDING, kuru_ending=POTENTIAL_KURU_POLITE_NEGATIVE_ENDING)
        else:
            verb_base = map_dictionary_to_e_ending(verb)

            if verb_class == VerbClass.GODAN:
                if formality == Formality.PLAIN:
                    return generate_nai_form(verb_base, verb_class, False)
                else:
                    return "{}{}".format(verb_base, MASU_NEGATIVE_NONPAST)
            elif verb_class == VerbClass.ICHIDAN:
                verb_base = splice_verb(verb, verb_class)
                verb_with_rare = "{}{}".format(verb_base, "られ")
                if formality == Formality.PLAIN:
                    return generate_nai_form(verb_with_rare, verb_class, False)
                else:
                    return "{}{}".format(verb_with_rare, MASU_NEGATIVE_NONPAST)
    
    def generate_imperative_form(self, verb, verb_class, formality):
        '''Generate the negative imperative form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative imperative form of the verb based on the formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                return handle_irregular_verb(verb, suru_ending="{}{}".format(SURU_ENDING, NA_PARTICLE), kuru_ending="{}{}".format(KURU_ENDING, NA_PARTICLE))
            else:
                nai_form = generate_nai_form(verb, verb_class, True)
                return "{}{}{}".format(nai_form, DE_PARTICLE, KUDASAI)
        else:
            if formality == Formality.PLAIN:
                return "{}{}".format(verb, NA_PARTICLE)
            else:
                return "{}{}{}".format(generate_nai_form(verb, verb_class, True), DE_PARTICLE, KUDASAI)
    
    def generate_provisional_form(self, verb, verb_class, formality):
        '''Generate the negative provisional form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative provisional form of the verb based on the formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if formality == Formality.PLAIN:
                if splice_verb(verb, verb_class, False) == SURU_ENDING:
                    return handle_irregular_verb(verb, append_stem_particle=True, suru_ending=PROVISIONAL_ICHIDAN_PLAIN_NEGATIVE_ENDING)
                else:
                    return "{}{}".format(KO_PARTICLE, PROVISIONAL_ICHIDAN_PLAIN_NEGATIVE_ENDING)
            else:
                intermediate_verb = handle_irregular_verb(verb, append_stem_particle=True, suru_ending=MASU_NEGATIVE_NONPAST, kuru_ending=MASU_NEGATIVE_NONPAST)
                return "{}{}{}".format(intermediate_verb, NA_PARTICLE, RA_PARTICLE)
        else:
            verb_stem = splice_verb(verb, verb_class)
            if verb_class == VerbClass.GODAN:
                verb_with_a_ending = map_dictionary_to_a_ending(verb)
                return "{}{}".format(verb_with_a_ending, PROVISIONAL_ICHIDAN_PLAIN_NEGATIVE_ENDING)
            elif verb_class == VerbClass.ICHIDAN:
                return "{}{}".format(verb_stem, PROVISIONAL_ICHIDAN_PLAIN_NEGATIVE_ENDING)
    

    def generate_causative_form(self, verb, verb_class, formality):
        '''Generate the negative causative form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative causative form of the verb based on the formality
        parameter
        '''
        if verb_class == VerbClass.IRREGULAR:
            if splice_verb(verb, verb_class, False) != SURU_ENDING:
                if formality == Formality.PLAIN: 
                    return generate_nai_form(CAUSATIVE_KURU_NEGATIVE_BASE, verb_class, False)
                else:
                    return "{}{}".format(CAUSATIVE_KURU_NEGATIVE_BASE, MASU_NEGATIVE_NONPAST)
        else:
            verb_stem = splice_verb(verb, verb_class)
            if verb_class == VerbClass.GODAN:
                modified_verb_stem = "{}{}".format(map_dictionary_to_a_ending(verb), SE_PARTICLE)
                if formality == Formality.PLAIN:
                    return generate_nai_form(modified_verb_stem, verb_class, False)
                else:
                    return "{}{}".format(modified_verb_stem, MASU_NEGATIVE_NONPAST)
            elif verb_class == VerbClass.ICHIDAN:
                modified_verb_stem = "{}{}{}".format(verb_stem, SA_PARTICLE, SE_PARTICLE)
                if formality == Formality.PLAIN:
                    return generate_nai_form(modified_verb_stem, verb_class, False)
                else:
                    return "{}{}".format(modified_verb_stem, MASU_NEGATIVE_NONPAST)

    def generate_passive_form(self, verb, verb_class, formality):
        '''Generate the negative passive form of the verb depending
        on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 

        Returns:
            str: negative passive form of the verb based on the formality
        parameter
        '''
        if verb_class == VerbClass.GODAN:
            verb_with_updated_ending = "{}{}".format(map_dictionary_to_a_ending(verb), RE_PARTICLE)
            if formality == Formality.PLAIN:
                return generate_nai_form(verb_with_updated_ending, verb_class, False)
            else:
                return "{}{}".format(verb_with_updated_ending, MASU_NEGATIVE_NONPAST)
        elif verb_class == VerbClass.ICHIDAN:
            modified_verb_stem = "{}{}{}".format(splice_verb(verb, verb_class), RA_PARTICLE, RE_PARTICLE)
            if formality == Formality.PLAIN:
                return generate_nai_form(modified_verb_stem, verb_class, False)
            else:
                return "{}{}".format(modified_verb_stem, MASU_NEGATIVE_NONPAST)



import unittest

from src.Utils import *
from src.constants.ParticleConstants import CHISAI_TSU_PARTICLE, KU_PARTICLE, GU_PARTICLE, SU_PARTICLE, U_PARTICLE, TSU_PARTICLE, TA_PARTICLE, DA_PARTICLE

from TestConstants import GodanVerbNomu, IchidanVerbTaberu, IrregularVerbSuru, IrregularVerbKuru


class UtilsTests(unittest.TestCase):
    def setUp(self):
        self.verb_class = VerbClass.GODAN

    def test_handle_irregular_verb_non_irregular_verb_ending(self):
        result = handle_irregular_verb(GodanVerbNomu.Verb)
        # TODO: refactor to throw error
        self.assertIsNone(result)

    # all below this must be godan verb
    def test_base_te_ta_form_CHISAI_TSU(self):
        verb = "使う"
        result = base_te_ta_form(verb, self.verb_class, TA_PARTICLE, DA_PARTICLE)
        self.assertEqual(result, "使った")
    
    def test_base_te_ta_form_KU_PARTICLE(self):
        verb = "聞く"
        result = base_te_ta_form(verb, self.verb_class, TA_PARTICLE, DA_PARTICLE)
        self.assertEqual(result, "聞いた")

    def test_base_te_ta_form_GU_PARTICLE(self):
        verb = "泳ぐ"
        result = base_te_ta_form(verb, self.verb_class, TA_PARTICLE, DA_PARTICLE)
        self.assertEqual(result, "泳いだ")

    def test_base_te_ta_form_SU_PARTICLE(self):
        verb = "話す"
        result = base_te_ta_form(verb, self.verb_class, TA_PARTICLE, DA_PARTICLE)
        self.assertEqual(result, "話した")

    def test_map_dict_form_to_different_ending_U_PARTICLE(self):
        verb = "使う"
        result = map_dictionary_to_a_ending(verb)
        self.assertEqual(result, "使わ")
   
    def test_map_dict_form_to_different_ending_TSU_PARTICLE(self):
        verb = "立つ"
        result = map_dictionary_to_a_ending(verb)
        self.assertEqual(result, "立た")

    def test_map_dict_form_to_different_ending_SU_PARTICLE(self):
        verb = "話す"
        result = map_dictionary_to_a_ending(verb)
        self.assertEqual(result, "話さ")

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(UtilsTests)
    unittest.TextTestRunner(verbosity=2).run(suite)


# PARTICLE CONSTANTS
A_PARTICLE = "あ"
E_PARTICLE = "え"
I_PARTICLE = "い"
O_PARTICLE = "お"
U_PARTICLE = "う"

KA_PARTICLE = "か"
KE_PARTICLE = "け"
KI_PARTICLE = "き"
KO_PARTICLE = "こ"
KU_PARTICLE = "く"

GA_PARTICLE = "が"
GE_PARTICLE = "げ"
GI_PARTICLE = "ぎ"
GO_PARTICLE = "ご"
GU_PARTICLE = "ぐ"

SA_PARTICLE = "さ"
SE_PARTICLE = "せ"
SHI_PARTICLE = "し"
SO_PARTICLE = "そ"
SU_PARTICLE = "す"

TA_PARTICLE = "た"
TE_PARTICLE = "て"
CHI_PARTICLE = "ち"
TO_PARTICLE = "と"
TSU_PARTICLE = "つ"

DA_PARTICLE = "だ"
DE_PARTICLE = "で"
DI_PARTICLE = "ぢ"
DO_PARTICLE = "ど"
DU_PARTICLE = "づ"

NA_PARTICLE = "な"
NE_PARTICLE = "ね"
NI_PARTICLE = "に"
NO_PARTICLE = "の"
NU_PARTICLE = "ぬ"

HA_PARTICLE = "は"
HE_PARTICLE = "へ"
HI_PARTICLE = "ひ"
HO_PARTICLE = "ほ"
HU_PARTICLE = "ふ"

BA_PARTICLE = "ば"
BE_PARTICLE = "べ"
BI_PARTICLE = "び"
BO_PARTICLE = "ぼ"
BU_PARTICLE = "ぶ"

PA_PARTICLE = "ぱ"
PE_PARTICLE = "ぺ"
PI_PARTICLE = "ぴ"
PO_PARTICLE = "ぽ"
PU_PARTICLE = "ぷ"

MA_PARTICLE = "ま"
ME_PARTICLE = "め"
MI_PARTICLE = "み"
MO_PARTICLE = "も"
MU_PARTICLE = "む"

YA_PARTICLE = "や"
YU_PARTICLE = "ゆ"
YO_PARTICLE = "よ"

RA_PARTICLE = "ら"
RE_PARTICLE = "れ"
RI_PARTICLE = ""
RO_PARTICLE = "ろ"
RU_PARTICLE = "る"

WA_PARTICLE = "わ"
WO_PARTICLE = "を"
N_PARTICLE = "ん"

CHISAI_TSU_PARTICLE = "っ"



from .constants.ParticleConstants import U_PARTICLE, KU_PARTICLE, GU_PARTICLE, SU_PARTICLE, TSU_PARTICLE, NU_PARTICLE, BU_PARTICLE, MU_PARTICLE, RU_PARTICLE

def containsJapaneseCharacters(verb):
    ''' Compute whether or not a Japanese verb contains any kanji characters

    Args:
        verb (str): Japanese verb in kana or kanji

    Returns:
        bool: True if kanji is found, false otherwise
    '''
    ranges = [
        # https://stackoverflow.com/questions/30069846/how-to-find-out-chinese-or-japanese-character-in-a-string-in-python
        {"from": ord(u"\u3300"), "to": ord(u"\u33ff")},         # compatibility ideographs
        {"from": ord(u"\ufe30"), "to": ord(u"\ufe4f")},         # compatibility ideographs
        {"from": ord(u"\uf900"), "to": ord(u"\ufaff")},         # compatibility ideographs
        {"from": ord(u"\U0002F800"), "to": ord(u"\U0002fa1f")}, # compatibility ideographs
        {'from': ord(u'\u3040'), 'to': ord(u'\u309f')},         # Japanese Hiragana
        {"from": ord(u"\u30a0"), "to": ord(u"\u30ff")},         # Japanese Katakana
        {"from": ord(u"\u2e80"), "to": ord(u"\u2eff")},         # cjk radicals supplement
        {"from": ord(u"\u4e00"), "to": ord(u"\u9fff")},
        {"from": ord(u"\u3400"), "to": ord(u"\u4dbf")},
        {"from": ord(u"\U00020000"), "to": ord(u"\U0002a6df")},
        {"from": ord(u"\U0002a700"), "to": ord(u"\U0002b73f")},
        {"from": ord(u"\U0002b740"), "to": ord(u"\U0002b81f")},
        {"from": ord(u"\U0002b820"), "to": ord(u"\U0002ceaf")}  # included as of Unicode 8.0
    ]

    for char in verb:
        if not any([range["from"] <= ord(char) <= range["to"] for range in ranges]):
            return False
    return True

def validateJapaneseVerbDecorator(func):
    def wrapper(self, verb, *args):
        if len(verb) < 2:
            print("here")
            raise Exception("Invalid Japanese Verb Length", len(verb), verb)

        if verb[-1:] not in [U_PARTICLE, KU_PARTICLE, GU_PARTICLE, SU_PARTICLE, TSU_PARTICLE, NU_PARTICLE, BU_PARTICLE, MU_PARTICLE, RU_PARTICLE]:
            raise Exception("Invalid Japanese Verb Ending Particle", verb[-1:])
        
        if not containsJapaneseCharacters(verb):
            raise Exception("Non-Japanese Character Found", verb)

        # assuming *args will always have the correct arguments because initial function call succeeded
        return func(self, verb, *args)
    return wrapper
    


import unittest

from src.JapaneseVerbFormGenerator import JapaneseVerbFormGenerator
from src.Decorators import *
from src.constants.EnumeratedTypes import VerbClass, Tense, Polarity
from src.constants.ParticleConstants import CHISAI_TSU_PARTICLE, KU_PARTICLE, GU_PARTICLE, SU_PARTICLE, U_PARTICLE, TSU_PARTICLE, TA_PARTICLE, DA_PARTICLE, PU_PARTICLE

from TestConstants import GodanVerbNomu, korean_with_japanese, english_with_japanese, verb_incorrect_particle_ending

class UtilsTests(unittest.TestCase):
    def setUp(self):
        self.verb_class = VerbClass.GODAN
        self.japaneseVerbFormGenerator = JapaneseVerbFormGenerator()

    def test_validateJapaneseVerbDecorator_InvalidLength(self):
        with self.assertRaises(Exception) as expectedException:
            self.japaneseVerbFormGenerator.generate_plain_form(KU_PARTICLE, VerbClass.GODAN, Tense.PAST, Polarity.NEGATIVE)
        self.assertEqual(str(expectedException.exception), '(\'Invalid Japanese Verb Length\', 1, \'' + KU_PARTICLE + '\')')
        
    # "Invalid Japanese Verb Ending Particle", verb[-1:]
    def test_validateJapaneseVerbDecorator_InvalidJapaneseEndingParticle(self):
        with self.assertRaises(Exception) as expectedException:
            self.japaneseVerbFormGenerator.generate_plain_form(verb_incorrect_particle_ending, VerbClass.GODAN, Tense.PAST, Polarity.NEGATIVE)
        self.assertEqual(str(expectedException.exception), '(\'Invalid Japanese Verb Ending Particle\', \'' + verb_incorrect_particle_ending[-1:] + '\')')

    def test_validateJapaneseVerbDecorator_ContainsKoreanCharacters(self):
        with self.assertRaises(Exception) as expectedException:
            self.japaneseVerbFormGenerator.generate_plain_form(korean_with_japanese, VerbClass.GODAN, Tense.PAST, Polarity.NEGATIVE)
        self.assertEqual(str(expectedException.exception), '(\'Non-Japanese Character Found\', \'' + korean_with_japanese + '\')')

    def test_validateJapaneseVerbDecorator_ContainsEnglishCharacters(self):
        with self.assertRaises(Exception) as expectedException:
            self.japaneseVerbFormGenerator.generate_plain_form(english_with_japanese, VerbClass.GODAN, Tense.PAST, Polarity.NEGATIVE)
        self.assertEqual(str(expectedException.exception), '(\'Non-Japanese Character Found\', \'' + english_with_japanese + '\')')

    def test_validateJapaneseVerbDecorator_ValidVerb(self):
        result = self.japaneseVerbFormGenerator.generate_plain_form(GodanVerbNomu.Verb, GodanVerbNomu.Verb_Class, Tense.NONPAST, Polarity.POSITIVE)
        self.assertEqual(result, GodanVerbNomu.Verb)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(UtilsTests)
    unittest.TextTestRunner(verbosity=2).run(suite)


from enum import Enum

class Formality(Enum):
    PLAIN = 1
    POLITE = 2

class Polarity(Enum):
    POSITIVE = 1
    NEGATIVE = 2

class Tense(Enum):
    PAST = 1
    NONPAST = 2

class VerbClass(Enum):
    GODAN = 1
    ICHIDAN = 2
    IRREGULAR = 3
    NONIRREGULAR = 4



from .EnumeratedTypes import Formality, Polarity, Tense, VerbClass

KUDASAI = "ください"

# IRREGULAR VERB ENDINGS
SURU_ENDING = "する"
KURU_ENDING = "くる"

# (POLITE) MASU FORM ENDINGS
MASU_POSITIVE_NONPAST = "ます"
MASU_NEGATIVE_NONPAST = "ません"
MASU_POSITIVE_PAST = "ました"
MASU_NEGATIVE_PAST = "ませんでした"

# (PLAIN) KATTA FORM ENDING
KATTA_ENDING = "かった"

# (PLAIN) NAI ENDING
NAI_ENDING = "ない"

# VOLITIONAL FORM ENDINGS
VOLITIONAL_POLITE_ENDING = "ましょう"
VOLITIONAL_POLITE_COPULA = "でしょう"
VOLITIONAL_PLAIN_COPULA = "だろう"
VOLITIONAL_ICHIDAN_PLAIN_ENDING = "よう"
VOLITIONAL_SURU_ENDING = "しよう"
VOLITIONAL_KURU_ENDING = "こよう"

# POTENTIAL FORM
POTENTIAL_POLITE_ICHIDAN_ENDING = "られます"
POTENTIAL_ICHIDAN_ENDING = "られる"
POTENTIAL_SURU_PLAIN_POSITIVE_ENDING = "できる"
POTENTIAL_SURU_PLAIN_NEGATIVE_ENDING = "できない"
POTENTIAL_SURU_POLITE_POSITIVE_ENDING = "できます"
POTENTIAL_SURU_POLITE_NEGATIVE_ENDING = "できません"
POTENTIAL_KURU_PLAIN_POSITIVE_ENDING = "こられる"
POTENTIAL_KURU_POLITE_POSITIVE_ENDING = "こられます"
POTENTIAL_KURU_PLAIN_NEGATIVE_ENDING = "こられない"
POTENTIAL_KURU_POLITE_NEGATIVE_ENDING = "こられません"

# PASSIVE FORM ENDINGS
PASSIVE_ICHIDAN_PLAIN_POSITIVE_ENDING = "られる"
PASSIVE_ICHIDAN_POLITE_POSITIVE_ENDING = "られます"
PASSIVE_ICHIDAN_PLAIN_NEGATIVE_ENDING = "られない"
PASSIVE_ICHIDAN_POLITE_NEGATIVE_ENDING = "られません"

# CAUSATIVE VERB ENDINGS
CAUSATIVE_PLAIN_SURU_ENDING = "させる"
CAUSATIVE_PLAIN_KURU_ENDING = "こさせる"
CAUSATIVE_KURU_NEGATIVE_BASE = "こさせ"

# IMPERATIVE VERB ENDINGS
IMPERATIVE_SURU_PLAIN_POSITIVE_ENDING = "しろ"
IMPERATIVE_KURU_PLAIN_POSITIVE_ENDING = "こい"

# PROVISIONAL VERB ENDINGS
PROVISIONAL_SURU_PLAIN_POSITIVE_ENDING = "すれば"
PROVISIONAL_SURU_POLITE_POSITIVE_ENDING = "しませば"
PROVISIONAL_KURU_PLAIN_POSITIVE_ENDING = "くれば"
PROVISIONAL_KURU_POLITE_POSITIVE_ENDING = "きませば"
PROVISIONAL_ICHIDAN_PLAIN_NEGATIVE_ENDING = "なければ"

# PASSIVE VERB ENDINGS
PASSIVE_SURU_PLAIN_POSITIVE_ENDING = "される"
PASSIVE_KURU_PLAIN_POSITIVE_ENDING = "こられる"



from src.constants.EnumeratedTypes import Formality, Polarity, Tense, VerbClass

# regex for kanji / kana: "[一-龯ぁ-んァ-ン]+"
# 飲む
# 
# 先秦兩漢

korean_with_japanese = "한국어처리기む"
english_with_japanese = "Heloむ"
verb_incorrect_particle_ending = "飲ま"


class GodanVerbNomu:
    # http://www.japaneseverbconjugator.com/VerbDetails.asp?txtVerb=%E9%A3%B2%E3%82%80
    Verb = "飲む" # plain positive nonpast
    Verb_Class = VerbClass.GODAN

    # Formal Verb Forms
    PolitePositiveNonpast = "飲みます"
    PolitePositivePast = "飲みました"
    PoliteNegativeNonpast = "飲みません"
    PoliteNegativePast = "飲みませんでした"

    # Plain Verb Forms
    PlainPositivePast = "飲んだ" # ta form
    PlainNegativeNonpast = "飲まない" # nai form
    PlainNegativePast = "飲まなかった" # katta form

    TeForm = "飲んで"

    # Conditional Verb Forms
    ConditionalPolite = "飲みましたら" # tara form
    ConditionalPlain = "飲んだら"

    # Volitional Verb Forms
    VolitionalPolitePositive = "飲みましょう"
    VolitionalPoliteNegative = "飲まないでしょう"
    VolitionalPlainPositive = "飲もう"
    VolitionalPlainNegative = "飲まないだろう"

    # Potential Verb Forms
    PotentialPlainPositive = "飲める"
    PotentialPlainNegative = "飲めない"
    PotentialPolitePositive = "飲めます"
    PotentialPoliteNegative = "飲めません"

    # Imperative Verb Forms
    ImperativePlainPositive = "飲め"
    ImperativePlainNegative= "飲むな"
    ImperativePolitePositive = "飲んでください"
    ImperativePoliteNegative = "飲まないでください"

    # Causative Verb Forms
    CausativePlainPositive = "飲ませる"
    CausativePlainNegative = "飲ませない"
    CausativePolitePositive = "飲ませます"
    CausativePoliteNegative = "飲ませません"

    # Passive Verb Forms
    PassivePlainPositive = "飲まれる"
    PassivePlainNegative = "飲まれない"
    PassivePolitePositive = "飲まれます"
    PassivePoliteNegative = "飲まれません"

    # Provisional Verb Forms
    ProvisionalPlainPositive = "飲めば"
    ProvisionalPlainNegative = "飲まなければ"


class IchidanVerbTaberu:
    # http://www.japaneseverbconjugator.com/VerbDetails.asp?txtVerb=%E9%A3%9F%E3%81%B9%E3%82%8B
    Verb = "食べる" # plain positive nonpast
    Verb_Class = VerbClass.ICHIDAN

    # Formal Verb Forms
    PolitePositiveNonpast = "食べます"
    PolitePositivePast = "食べました"
    PoliteNegativeNonpast = "食べません"
    PoliteNegativePast = "食べませんでした"

    # Plain Verb Forms
    PlainPositivePast = "食べた" # ta form
    PlainNegativeNonpast = "食べない" # nai form
    PlainNegativePast = "食べなかった" # katta form

    TeForm = "食べて"

    # Conditional Verb Forms
    ConditionalPolite = "食べましたら" # tara form
    ConditionalPlain = "食べたら" # tara form

    # Volitional Verb Forms
    VolitionalPlainPositive = "食べよう"
    VolitionalPlainNegative = "食べないだろう"
    VolitionalPolitePositive = "食べましょう"
    VolitionalPoliteNegative = "食べないでしょう"

    # Potential Verb Forms
    PotentialPlainPositive = "食べられる"
    PotentialPlainNegative = "食べられない"
    PotentialPolitePositive = "食べられます"
    PotentialPoliteNegative = "食べられません"

    # Imperative Verb Forms
    ImperativePlainPositive = "食べろ"
    ImperativePlainNegative= "食べるな"
    ImperativePolitePositive = "食べてください"
    ImperativePoliteNegative = "食べないでください"

    # Causative Verb Forms
    CausativePlainPositive = "食べさせる"
    CausativePlainNegative = "食べさせない"
    CausativePolitePositive = "食べさせます"
    CausativePoliteNegative = "食べさせません"

    # Passive Verb Forms
    PassivePlainPositive = "食べられる"
    PassivePlainNegative = "食べられない"
    PassivePolitePositive = "食べられます"
    PassivePoliteNegative = "食べられません"

    # Provisional Verb Forms
    ProvisionalPlainPositive = "食べれば"
    ProvisionalPlainNegative = "食べなければ"


class IrregularVerbSuru: 
    # http://www.japaneseverbconjugator.com/Suru.asp
    Verb = "勉強する" # plain positive nonpast
    Verb_Class = VerbClass.IRREGULAR

    # Formal Verb Forms
    PolitePositiveNonpast = "勉強します"
    PolitePositivePast = "勉強しました"
    PoliteNegativeNonpast = "勉強しません"
    PoliteNegativePast = "勉強しませんでした"

    # Plain Verb Forms
    PlainPositivePast = "勉強した" # ta form
    PlainNegativeNonpast = "勉強しない" # nai form
    PlainNegativePast = "勉強しなかった" # katta form

    TeForm = "勉強して"

    # Conditional Verb Forms
    ConditionalPlainPositive = "勉強したら" # tara form
    ConditionalPolitePositive = "勉強しましたら" # tara form
    ConditionalPlainNegative = "勉強しなかったら" # tara form
    ConditionalPoliteNegative = "勉強しませんでしたら" # tara form

    # Volitional Verb Forms
    VolitionalPlainPositive = "勉強しよう"
    VolitionalPolitePositive = "勉強しましょう "
    VolitionalPlainNegative = "勉強しないだろう"
    VolitionalPoliteNegative = "勉強しないでしょう"

    # VolitionalPlainPositivePast = "勉強したろう"
    # VolitionalPolitePositivePast = "勉強しましたろう"
    # VolitionalPlainNegativePast = "勉強しなかっただろう"
    # VolitionalPoliteNegativePast = "勉強しなかたでしょう"

    # Potential Verb Forms
    PotentialPlainPositive = "勉強できる"
    PotentialPlainNegative = "勉強できない"
    PotentialPolitePositive = "勉強できます"
    PotentialPoliteNegative = "勉強できません"

    # Imperative Verb Forms
    ImperativePlainPositive = "勉強しろ"
    ImperativePlainNegative= "勉強するな"
    ImperativePolitePositive = "勉強してください"
    ImperativePoliteNegative = "勉強しないでください"

    # Causative Verb Forms
    CausativePlainPositive = "勉強させる"

    # Passive Verb Forms
    PassivePlainPositive = "勉強される"

    # Provisional Verb Forms
    ProvisionalPlainPositive = "勉強すれば"
    ProvisionalPlainNegative = "勉強しなければ"
    ProvisionalPolitePositive = "勉強しませば"
    ProvisionalPoliteNegative = "勉強しませんなら"


class IrregularVerbKuru: 
    # http://www.japaneseverbconjugator.com/Kuru.asp
    Verb = "くる" # plain positive nonpast
    Verb_Class = VerbClass.IRREGULAR

    # Formal Verb Forms
    PolitePositiveNonpast = "きます"
    PolitePositivePast = "きました"
    PoliteNegativeNonpast = "きません"
    PoliteNegativePast = "きませんでした"

    # Plain Verb Forms
    PlainPositivePast = "きた" # ta form
    PlainNegativeNonpast = "こない" # nai form
    PlainNegativePast = "こなかった" # katta form

    TeForm = "きて"

    # Conditional Verb Forms
    ConditionalPlainPositive = "きたら" # tara form
    ConditionalPolitePositive = "きましたら" # tara form
    ConditionalPlainNegative = "こなかったら" # tara form
    ConditionalPoliteNegative = "きませんでしたら" # tara form

    # Volitional Verb Forms
    VolitionalPolitePositive = "きましょう"
    VolitionalPoliteNegative = "こないでしょう"
    VolitionalPlainPositive = "こよう"
    VolitionalPlainNegative = "こないだろう"

    # VolitionalPlainPositivePast = "きたろう"
    # VolitionalPolitePositivePast = "きたでしょう"
    # VolitionalPlainNegativePast = "こなかっただろう"
    # VolitionalPoliteNegativePast = "こなかったでしょう"

    # Potential Verb Forms
    PotentialPlainPositive = "こられる"
    PotentialPlainNegative = "こられない"
    PotentialPolitePositive = "こられます"
    PotentialPoliteNegative = "こられません"

    # Imperative Verb Forms
    ImperativePlainPositive = "こい"
    ImperativePlainNegative= "くるな"
    ImperativePolitePositive = "きてください"
    ImperativePoliteNegative = "こないでください"

    # Causative Verb Forms
    CausativePlainPositive = "こさせる"
    CausativePlainNegative = "こさせない"
    CausativePolitePositive = "こさせます"
    CausativePoliteNegative = "こさせません"

    # Passive Verb Forms
    PassivePlainPositive = "こられる"

    # Provisional Verb Forms
    ProvisionalPlainPositive = "くれば"
    ProvisionalPlainNegative = "こなければ"
    ProvisionalPolitePositive = "きませば"
    ProvisionalPoliteNegative = "きませんなら"


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='JapaneseVerbConjugator',
    version='1.0.0',
    author='Jacob Shiohira',
    author_email='shiohira.jacob@gmail.com',
    description='Japanese verb Conjugator',
    long_description=long_description,
    url='https://github.com/jShiohaha/JapaneseVerbConjugator/',
    packages=setuptools.find_packages(),
    install_requires=[
          'romkan',
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
    ],
)


# Local modules
from .constants.EnumeratedTypes import Polarity

from .Decorators import validateJapaneseVerbDecorator
from .PositiveVerbFormGenerator import PositiveVerbForms
from .NegativeVerbFormGenerator import NegativeVerbForms


class JapaneseVerbFormGenerator():
    def __init__(self):
        self.positiveVerbForms = PositiveVerbForms()
        self.negativeVerbForms = NegativeVerbForms()

    @validateJapaneseVerbDecorator
    def generate_plain_form(self, verb, verb_class, tense, polarity):
        '''Generate the plain form of the verb depending on the tense and 
        polarity.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: plain form of the verb based on the tense and polarity
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_plain_form(verb, verb_class, tense)
        return self.negativeVerbForms.generate_plain_form(verb, verb_class, tense)

    @validateJapaneseVerbDecorator
    def generate_polite_form(self, verb, verb_class, tense, polarity):
        '''Generate the polite form of the verb depending on the tense and 
        polarity.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            tense (enum): Tense Enum representing the tense for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: polite form of the verb based on the tense and polarity
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_polite_form(verb, verb_class, tense)
        return self.negativeVerbForms.generate_polite_form(verb, verb_class, tense)

    @validateJapaneseVerbDecorator
    def generate_te_form(self, verb, verb_class):
        '''Utilize base_te_ta_form function to generate the -te form 
        of the verb

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs

        Returns:
            str: -te form of the verb
        '''
        return self.positiveVerbForms.generate_te_form(verb, verb_class)

    @validateJapaneseVerbDecorator
    def generate_conditional_form(self, verb, verb_class, formality, polarity):
        '''Generate the conditional form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: conditional form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_conditional_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_conditional_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_volitional_form(self, verb, verb_class, formality, polarity):
        '''Generate the volitional form of the verb depending on the formality. 

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: volitional form of the verb based on the formality and polarity 
        parameters
        '''        
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_volitional_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_volitional_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_potential_form(self, verb, verb_class, formality, polarity):
        '''Generate the potential form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: potential form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_potential_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_potential_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_imperative_form(self, verb, verb_class, formality, polarity):
        '''Generate the imperative form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: imperative form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_imperative_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_imperative_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_provisional_form(self, verb, verb_class, formality, polarity):
        '''Generate the provisional form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: provisional form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_provisional_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_provisional_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_causative_form(self, verb, verb_class, formality, polarity):
        '''Generate the causative form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: causative form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_causative_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_causative_form(verb, verb_class, formality)

    @validateJapaneseVerbDecorator
    def generate_passive_form(self, verb, verb_class, formality, polarity):
        '''Generate the passive form of the verb depending on the formality.

        Args:
            verb (str): Japanese verb in kana, might contain kanji
            verb_class (enum): VerbClass Enum representing the verb class
                to which the verb belongs
            formality (enum): Formality Enum representing the formality class
                for the conjugated verb 
            polarity (enum): Polarity Enum representing the polarity for the 
                conjugated verb

        Returns:
            str: passive form of the verb based on the formality and polarity 
        parameters
        '''
        if polarity == Polarity.POSITIVE:
            return self.positiveVerbForms.generate_passive_form(verb, verb_class, formality)
        return self.negativeVerbForms.generate_passive_form(verb, verb_class, formality)



import unittest

from src.JapaneseVerbFormGenerator import *
from src.constants.EnumeratedTypes import Polarity, Formality, VerbClass, Tense

from TestConstants import GodanVerbNomu, IchidanVerbTaberu, IrregularVerbSuru, IrregularVerbKuru

# https://github.com/audreyr/how-to/blob/master/python/use_coverage_with_unittest.rst
# https://github.com/audreyr/how-to/blob/master/python/use_coverage_with_unittest.rst#user-content-set-up-coveralls

class ParametrizedTestCase(unittest.TestCase):
    """ TestCase classes that want to be parametrized should
        inherit from this class

        Borrowed from https://eli.thegreenplace.net/2011/08/02/python-unit-testing-parametrized-test-cases/
    """
    def __init__(self, methodName='runTest', param=None):
        super(ParametrizedTestCase, self).__init__(methodName)
        self.japaneseVerbFormGenerator = JapaneseVerbFormGenerator()
        self.param = param

    @staticmethod
    def parametrize(testcase_klass, param=None):
        """ Create a suite containing all tests taken from the given
            subclass, passing them the parameter 'param'
        """
        testloader = unittest.TestLoader()
        testnames = testloader.getTestCaseNames(testcase_klass)
        suite = unittest.TestSuite()
        for name in testnames:
            suite.addTest(testcase_klass(name, param=param))
        return suite

# ---------------------------------------------------------- #
#                    Positive Verb Form Tests                #
# ---------------------------------------------------------- #
class TestPositiveVerbForms(ParametrizedTestCase):
    ''' Tests for each positive verb conjugation form considering
    the permutations of tense and formality. Skipped Tests for 
    Irregular Positive Verbs include:

    - CausativePolitePositive
    - ConditionalPlain
    - ConditionalPolite
    - PassivePolitePositive
    - VolitionalPolitePositive

    This structure of class based test cases and inheriting from 
    ParametrizedTestCase allows testing of different verbs across
    different verb classes. To test a different verb, simply update
    the self.param arg passed to ParametrizedTestCase. 
    '''
    def setUp(self):
        self.polarity = Polarity.POSITIVE
        self.verb = self.param
    
    # Polite Verb Forms Test
    def test_polite_positive_nonpast(self):
        result = self.japaneseVerbFormGenerator.generate_polite_form(self.verb.Verb, self.verb.Verb_Class, Tense.NONPAST, self.polarity)
        self.assertEqual(result, self.verb.PolitePositiveNonpast)

    def test_polite_positive_past(self):
        result = self.japaneseVerbFormGenerator.generate_polite_form(self.verb.Verb, self.verb.Verb_Class, Tense.PAST, self.polarity)
        self.assertEqual(result, self.verb.PolitePositivePast)

    # Plain Verb Forms Test
    def test_plain_positive_nonpast(self):
        result = self.japaneseVerbFormGenerator.generate_plain_form(self.verb.Verb, self.verb.Verb_Class, Tense.NONPAST, self.polarity)
        self.assertEqual(result, self.verb.Verb)

    def test_plain_positive_past (self):
        result = self.japaneseVerbFormGenerator.generate_plain_form(self.verb.Verb, self.verb.Verb_Class, Tense.PAST, self.polarity)
        self.assertEqual(result, self.verb.PlainPositivePast)

    # Te Verb Form Test
    def test_te_form(self):
        result = self.japaneseVerbFormGenerator.generate_te_form(self.verb.Verb, self.verb.Verb_Class)
        self.assertEqual(result, self.verb.TeForm)

    # Conditional Verb Forms Test
    def test_conditional_plain(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_conditional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ConditionalPlain)

    def test_conditional_polite(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_conditional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ConditionalPolite)

    # Volitional Verb Forms Test
    def test_volitional_plain_positive(self):
        # if self.verb.Verb_Class == VerbClass.IRREGULAR:
        #     self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_volitional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.VolitionalPlainPositive)

    def test_volitional_polite_positive(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_volitional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.VolitionalPolitePositive)

    # Potential Verb Forms Test
    def test_potential_plain_positive(self):
        result = self.japaneseVerbFormGenerator.generate_potential_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.PotentialPlainPositive)

    def test_potential_polite_positive(self):
        result = self.japaneseVerbFormGenerator.generate_potential_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.PotentialPolitePositive)

    # Imperative Verb Forms Test
    def test_imperative_plain_positive(self):
        result = self.japaneseVerbFormGenerator.generate_imperative_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ImperativePlainPositive)

    def test_imperative_polite_positive(self):
        result = self.japaneseVerbFormGenerator.generate_imperative_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ImperativePolitePositive)

    # Provisional Verb Forms Test
    def test_provisional_plain_positive(self):
        result = self.japaneseVerbFormGenerator.generate_provisional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ProvisionalPlainPositive)

    def test_provisional_polite_positive(self):
        if self.verb.Verb_Class is not VerbClass.IRREGULAR:
            self.skipTest("Not Required for Non-Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_provisional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ProvisionalPolitePositive)

    # Causative Verb Forms Test
    def test_causative_plain_positive(self):
        result = self.japaneseVerbFormGenerator.generate_causative_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.CausativePlainPositive)

    def test_causative_polite_positive(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_causative_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.CausativePolitePositive)

    # Passive Verb Forms Test
    def test_passive_plain_positive(self):
        result = self.japaneseVerbFormGenerator.generate_passive_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.PassivePlainPositive)

    def test_passive_polite_positive(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not Required for Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_passive_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.PassivePolitePositive)

# ---------------------------------------------------------- #
#                    Negative Verb Form Tests                #
# ---------------------------------------------------------- #
class TestNegativeVerbForms(ParametrizedTestCase):
    ''' Tests for each negative verb conjugation form considering
    the permutations of tense and formality. Skipped Tests for 
    Irregular Negative Verbs include:

    - CausativePlainNegative for する verbs
    - CausativePoliteNegative for する verbs
    - PassivePlainNegative
    - PassivePoliteNegative
    - ProvisionalPoliteNegative

    This structure of class based test cases and inheriting from 
    ParametrizedTestCase allows testing of different verbs across
    different verb classes. To test a different verb, simply update
    the self.param arg passed to ParametrizedTestCase. 
    '''
    def setUp(self):
        self.polarity = Polarity.NEGATIVE
        self.verb = self.param

    # Formal Verb Forms Test
    def test_polite_negative_nonpast(self):
        result = self.japaneseVerbFormGenerator.generate_polite_form(self.verb.Verb, self.verb.Verb_Class, Tense.NONPAST, self.polarity)
        self.assertEqual(result, self.verb.PoliteNegativeNonpast)

    def test_polite_negative_past(self):
        result = self.japaneseVerbFormGenerator.generate_polite_form(self.verb.Verb, self.verb.Verb_Class, Tense.PAST, self.polarity)
        self.assertEqual(result, self.verb.PoliteNegativePast)

    # Plain Verb Forms Test
    def test_plain_negative_nonpast (self):
        result = self.japaneseVerbFormGenerator.generate_plain_form(self.verb.Verb, self.verb.Verb_Class, Tense.NONPAST, self.polarity)
        self.assertEqual(result, self.verb.PlainNegativeNonpast)

    def test_plain_negative_past (self):
        result = self.japaneseVerbFormGenerator.generate_plain_form(self.verb.Verb, self.verb.Verb_Class, Tense.PAST, self.polarity)
        self.assertEqual(result, self.verb.PlainNegativePast)

    # Conditional Verb Forms Test
    def test_conditional_plain(self):
        if self.verb.Verb_Class is not VerbClass.IRREGULAR:
            self.skipTest("Not Required for Non-Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_conditional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ConditionalPlainNegative)

    def test_conditional_polite(self):
        if self.verb.Verb_Class is not VerbClass.IRREGULAR:
            self.skipTest("Not Required for Non-Irregular Verbs")
        result = self.japaneseVerbFormGenerator.generate_conditional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ConditionalPoliteNegative)

    # Volitional Verb Forms Test
    def test_volitional_plain_negative(self):
        result = self.japaneseVerbFormGenerator.generate_volitional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.VolitionalPlainNegative)

    def test_volitional_polite_negative(self):
        result = self.japaneseVerbFormGenerator.generate_volitional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.VolitionalPoliteNegative)

    # Potential Verb Forms Test
    def test_potential_plain_negative(self):
        result = self.japaneseVerbFormGenerator.generate_potential_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.PotentialPlainNegative)

    def test_potential_polite_negative(self):
        result = self.japaneseVerbFormGenerator.generate_potential_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.PotentialPoliteNegative)

    # Imperative Verb Forms Test
    def test_imperative_plain_negative(self):
        result = self.japaneseVerbFormGenerator.generate_imperative_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ImperativePlainNegative)

    def test_imperative_polite_negative(self):
        result = self.japaneseVerbFormGenerator.generate_imperative_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ImperativePoliteNegative)

    # Provisional Verb Forms Test
    def test_provisional_plain_negative(self):
        result = self.japaneseVerbFormGenerator.generate_provisional_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.ProvisionalPlainNegative)

    def test_provisional_polite_negative(self):
        if self.verb.Verb_Class is not VerbClass.IRREGULAR:
            self.skipTest("Not required for non-irregular verbs")
        result = self.japaneseVerbFormGenerator.generate_provisional_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.ProvisionalPoliteNegative)

    # Causative Verb Forms Test
    def test_causative_plain_negative(self):
        if self.verb.Verb[-2:] == "する":
            self.skipTest("Not required for する verbs")
        result = self.japaneseVerbFormGenerator.generate_causative_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.CausativePlainNegative)

    def test_causative_polite_negative(self):
        if self.verb.Verb[-2:] == "する":
            self.skipTest("Not required for する verbs")
        result = self.japaneseVerbFormGenerator.generate_causative_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.CausativePoliteNegative)

    # Passive Verb Forms Test
    def test_passive_plain_negative(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not required for irregular verbs")
        result = self.japaneseVerbFormGenerator.generate_passive_form(self.verb.Verb, self.verb.Verb_Class, Formality.PLAIN, self.polarity)
        self.assertEqual(result, self.verb.PassivePlainNegative)

    def test_passive_polite_negative(self):
        if self.verb.Verb_Class == VerbClass.IRREGULAR:
            self.skipTest("Not required for irregular verbs")
        result = self.japaneseVerbFormGenerator.generate_passive_form(self.verb.Verb, self.verb.Verb_Class, Formality.POLITE, self.polarity)
        self.assertEqual(result, self.verb.PassivePoliteNegative)

# ---------------------------------------------------------- #
#     Register Positive and Negative Verb Form Test Suites   #
# ---------------------------------------------------------- #
def create_positive_verb_form_suite(suite):
    ''' Adds all positive verb form generator tests for each specified method. 
        The current verbs tested are:
         
        1. Godan Verb [Nomu / のむ]
        2. Ichidan Verb [Taberu / たベる]
        3. Irregular Verb [Suru / する]
        4. Irregular Verb [Kuru / くる]
    '''
    suite.addTest(ParametrizedTestCase.parametrize(TestPositiveVerbForms, param=GodanVerbNomu))
    suite.addTest(ParametrizedTestCase.parametrize(TestPositiveVerbForms, param=IchidanVerbTaberu))
    suite.addTest(ParametrizedTestCase.parametrize(TestPositiveVerbForms, param=IrregularVerbSuru))
    suite.addTest(ParametrizedTestCase.parametrize(TestPositiveVerbForms, param=IrregularVerbKuru))

    return suite

def create_negative_verb_form_suite(suite):
    ''' Adds all negative verb form generator tests for each specified method. 
        The current verbs tested are:

        1. Godan Verb [Nomu / のむ]
        2. Ichidan Verb [Taberu / たベる]
        3. Irregular Verb [Suru / する]
        4. Irregular Verb [Kuru / くる]
    '''
    suite.addTest(ParametrizedTestCase.parametrize(TestNegativeVerbForms, param=GodanVerbNomu))
    suite.addTest(ParametrizedTestCase.parametrize(TestNegativeVerbForms, param=IchidanVerbTaberu))
    suite.addTest(ParametrizedTestCase.parametrize(TestNegativeVerbForms, param=IrregularVerbSuru))
    suite.addTest(ParametrizedTestCase.parametrize(TestNegativeVerbForms, param=IrregularVerbKuru))

    return suite

if __name__ == '__main__':
    suite = unittest.TestSuite()
    create_positive_verb_form_suite(suite)
    create_negative_verb_form_suite(suite)

    unittest.TextTestRunner(verbosity=2).run(suite)


# Local modules
from .constants.ParticleConstants import *
from .constants.VerbEndingConstants import *
from .constants.EnumeratedTypes import Formality, Polarity, Tense, VerbClass

# External Libraries
import romkan

# ---------------------------------------------------------- #
#                UTIL VERB GENERATOR FUNCTIONS               #
# ---------------------------------------------------------- #
def splice_verb(verb, verb_class, should_return_stem=True):
    '''Split Japense verb between stem and ending particle(s). The number of ending
    particles returned depends on the verb class (i.e. godan / ichidan will return one
    particle while irregular verbs will return two particles)

    Args:
        verb (str): Japanese verb in kanji and/or kana
        verb_class (enum): VerbClass enum representing the Japanese verb class

    Returns:
        str: Verb stem or particle endings based on the value of should_return_stem
    '''
    num_ending_particles = 1
    if verb_class == VerbClass.IRREGULAR:
        num_ending_particles = 2
    if should_return_stem:
        return verb[:-1*num_ending_particles] 
    return verb[-1*num_ending_particles:] 

def handle_irregular_verb(verb, append_stem_particle=False, suru_ending=None, kuru_ending=None):
    ''' Handles irregular verb conjugations depending on suru or kuru verb type.
    Isolates logic of irregular verbs.
    
    Args:
        verb (str): Japanese verb in kana, might contain kanji
        append_stem_particle (bool): verb base particle depends on if the verb is a 
            suru or kuru verb. し particle is appended to suru verbs and き is appended
            to kuru verbs. Not all conjugations require these particles to be appended
            to the verb stem.
        suru_ending (:obj: str, optional): suru verb ending based on the conjugation form.
            Defaults to None.
        kuru_ending (:obj: str, optional): kuru verb ending based on the conjugation form.
            Defaults to None.

    Returns:
        str: irregular verb with appropriate particles and ending attached depending
            on verb conjugation
    '''
    particle_ending = splice_verb(verb, VerbClass.IRREGULAR, False)
    if particle_ending not in [SURU_ENDING, KURU_ENDING]:
        return None

    verb_stem = splice_verb(verb, VerbClass.IRREGULAR)
    ending = ""
    if particle_ending == SURU_ENDING:
        if append_stem_particle:
            ending = SHI_PARTICLE
        if suru_ending is not None:
            ending = "{}{}".format(ending, suru_ending)
    else:
        if append_stem_particle:
            ending = KI_PARTICLE
        if kuru_ending is not None:
            ending = "{}{}".format(ending, kuru_ending)
    return "{}{}".format(verb_stem, ending)

def generate_nai_form(verb, verb_class, is_regular_nai):
    ''' Generates the nai form of a Japanese verb

    Args:
        verb (str): Japanese verb in kana, might contain kanji
        verb_class (enum): VerbClass Enum representing the verb class
            to which the verb belongs
        is_regular_nai (bool): indicates whether to attach nai ending directly
            to verb param or generate the explicit nai form of a verb. Some
            verb conjugations require the nai ending without calculating the 
            nai form

    Returns:
        str: nai ending attached to verb param
    '''
    verb_stem = splice_verb(verb, verb_class)
    ending = NAI_ENDING

    if not is_regular_nai:
        return "{}{}".format(verb, ending)
    if verb_class == VerbClass.IRREGULAR:
        if splice_verb(verb, verb_class, False) == SURU_ENDING:
            ending = "{}{}".format(SHI_PARTICLE, ending)
        else:
            ending = "{}{}".format(KO_PARTICLE, ending)
    else:
        if verb_class == VerbClass.GODAN:
            verb_stem = map_dictionary_to_a_ending(verb)
    return "{}{}".format(verb_stem, ending)

def base_te_ta_form(verb, verb_class, *endings):
    ''' Handle the formation of the -te / -ta form for verbs belonging to
    any verb class. Logic for both forms follows similar logic but differs
    between (-te, -de) and (-ta, -da) based on the last particle of a Godan 
    verb.

    Args:
        verb (str): Japanese verb in kana, might contain kanji
        verb_class (enum): VerbClass Enum representing the verb class
            to which the verb belongs
        *endings: Variable length argument list. Must be in the form (te, de)
        or (ta, da) 
        
        TODO... reformat this logic for *endings

    Returns:
        str: The verb stem plus the -te / -ta particle depending on the 
        verb class. Defaults to None. 
    '''
    if verb_class == VerbClass.IRREGULAR:
        return handle_irregular_verb(verb, True, endings[0], endings[0])
    else:
        verb_stem = splice_verb(verb, verb_class)
        verb_ending = ""
        if verb_class == VerbClass.ICHIDAN:
            verb_ending = endings[0]
        elif verb_class == VerbClass.GODAN:
            last_kana = splice_verb(verb, verb_class, False)
            
            if last_kana in [RU_PARTICLE, TSU_PARTICLE, U_PARTICLE]:
                verb_ending = "{}{}".format(CHISAI_TSU_PARTICLE, endings[0])
            elif last_kana in [BU_PARTICLE, MU_PARTICLE, NU_PARTICLE]:
                verb_ending = "{}{}".format(N_PARTICLE, endings[1])
            elif last_kana in [KU_PARTICLE]:
                verb_ending = "{}{}".format(I_PARTICLE, endings[0])
            elif last_kana in [GU_PARTICLE]:
                verb_ending = "{}{}".format(I_PARTICLE, endings[1])
            elif last_kana in [SU_PARTICLE]:
                verb_ending = "{}{}".format(SHI_PARTICLE, endings[0])
        return "{}{}".format(verb_stem, verb_ending)

def map_dictionary_to_a_ending(verb):
    ''' Generates Godan verb stem with corresponding -a particle attached

    Args:
        verb (str): Japanese verb in kana, might contain kanji

    Returns:
        str: verb stem with the correct -a particle attached (Godan verbs only)
    '''
    return map_dict_form_to_different_ending(verb, 'a', WA_PARTICLE, TA_PARTICLE, SA_PARTICLE)

def map_dictionary_to_e_ending(verb):
    '''Generates Godan verb stem with corresponding -e particle attached

    Args:
        verb (str): Japanese verb in kana, might contain kanji

    Returns:
        str: verb stem with the correct -e particle attached (Godan verbs only)
    '''
    return map_dict_form_to_different_ending(verb, 'e', E_PARTICLE, TE_PARTICLE, SE_PARTICLE)

def map_dictionary_to_i_ending(verb):
    '''Generates Godan verb stem with corresponding -i particle attached

    Args:
        verb (str): Japanese verb in kana, might contain kanji

    Returns:
        str: verb stem with the correct -i particle attached (Godan verbs only)
    '''
    return map_dict_form_to_different_ending(verb, 'i', I_PARTICLE, CHI_PARTICLE, SHI_PARTICLE)

def map_dictionary_to_o_ending(verb):
    '''Generates Godan verb stem with corresponding -o particle attached

    Args:
        verb (str): Japanese verb in kana, might contain kanji

    Returns:
        str: verb stem with the correct -o particle attached (Godan verbs only)
    '''
    return map_dict_form_to_different_ending(verb, 'o', O_PARTICLE, TO_PARTICLE, SO_PARTICLE)

def map_dict_form_to_different_ending(verb, romaji_ending, *special_endings):
    '''Generates Godan verb stem and computes the correct particle to attach based on the
    verb's last kana 

    Args:
        verb (str): Japanese verb in kana, might contain kanji
        romaji_ending (str): target sound of the particle to append to the verb
        *special_endings: Variable length argument list. Based on the target Godan particle 
        class (-a, -e, -i, -o). Order of particles is -u / -tsu / -su.

    Returns:
        str: verb stem with the correct particle attached depending on the last kana particle
    of the Godan verb
    '''
    last_kana = splice_verb(verb, VerbClass.GODAN, False)
    verb_stem = splice_verb(verb, VerbClass.GODAN)

    if last_kana == U_PARTICLE:
        return "{}{}".format(verb_stem, special_endings[0])
    elif last_kana == TSU_PARTICLE:
        return "{}{}".format(verb_stem, special_endings[1])
    elif last_kana == SU_PARTICLE:
        return "{}{}".format(verb_stem, special_endings[2])
    else:
        transformed_last_kana_as_romaji = "{}{}".format(romkan.to_roma(last_kana)[:-1], romaji_ending)
        return "{}{}".format(verb_stem, romkan.to_hiragana(transformed_last_kana_as_romaji))
